class ActionKey:
    TURN_OFF = "turn_off"
    RESTART = "restart"


class ActionName:
    TURN_OFF = "Turn off"
    RESTART = "Restart"


class SensorKey:
    NETWORK_INTERFACE = "network_interface"
    MAC_ADDRESS = "mac_address"
    WAKE_ON_LAN = "wake_on_lan"
    MACHINE_TYPE = "machine_type"
    HOSTNAME = "hostname"
    OS_NAME = "os_name"
    OS_VERSION = "os_version"
    OS_ARCHITECTURE = "os_architecture"
    DEVICE_NAME = "device_name"
    DEVICE_MODEL = "device_model"
    MANUFACTURER = "manufacturer"
    SERIAL_NUMBER = "serial_number"
    CPU_NAME = "cpu_name"
    CPU_CORES = "cpu_cores"
    CPU_HARDWARE = "cpu_hardware"
    CPU_MODEL = "cpu_model"
    TOTAL_MEMORY = "total_memory"
    FREE_MEMORY = "free_memory"
    CPU_LOAD = "cpu_load"
    FREE_DISK_SPACE = "free_disk_space"
    TEMPERATURE = "temperature"
    PROCESSES = "processes"


class SensorName:
    NETWORK_INTERFACE = "Network interface"
    MAC_ADDRESS = "MAC address"
    WAKE_ON_LAN = "Wake on LAN"
    MACHINE_TYPE = "Machine type"
    HOSTNAME = "Hostname"
    OS_NAME = "OS name"
    OS_VERSION = "OS version"
    OS_ARCHITECTURE = "OS architecture"
    DEVICE_NAME = "Device name"
    DEVICE_MODEL = "Device model"
    MANUFACTURER = "Manufacturer"
    SERIAL_NUMBER = "Serial number"
    CPU_NAME = "CPU name"
    CPU_CORES = "CPU cores"
    CPU_HARDWARE = "CPU hardware"
    CPU_MODEL = "CPU model"
    TOTAL_MEMORY = "Total memory"
    FREE_MEMORY = "Free memory"
    CPU_LOAD = "CPU load"
    FREE_DISK_SPACE = "Free disk space"
    TEMPERATURE = "Temperature"
    PROCESSES = "Processes"



from ..collection import Collection
from ..command import ActionCommand, SensorCommand
from ..sensor import BinarySensor, NumberSensor, TextSensor
from .const import ActionKey, ActionName, SensorKey, SensorName

linux = Collection(
    "Linux",
    [
        ActionCommand(
            "/sbin/shutdown -h now",
            ActionName.TURN_OFF,
            ActionKey.TURN_OFF,
        ),
        ActionCommand(
            "/sbin/shutdown -r now",
            ActionName.RESTART,
            ActionKey.RESTART,
        ),
    ],
    [
        # TODO: OS_ARCHITECTURE
        SensorCommand(
            "/sbin/route -n | awk '/^0.0.0.0/ {print $NF}'",
            sensors=[
                TextSensor(
                    SensorName.NETWORK_INTERFACE,
                    SensorKey.NETWORK_INTERFACE,
                )
            ],
        ),
        SensorCommand(
            "cat /sys/class/net/&{network_interface}/address",
            sensors=[
                TextSensor(
                    SensorName.MAC_ADDRESS,
                    SensorKey.MAC_ADDRESS,
                )
            ],
        ),
        SensorCommand(
            "cat /sys/class/net/&{network_interface}/device/power/wakeup 2>/dev/null",
            sensors=[
                BinarySensor(
                    SensorName.WAKE_ON_LAN,
                    SensorKey.WAKE_ON_LAN,
                )
            ],
        ),
        SensorCommand(
            "uname -n",
            sensors=[
                TextSensor(
                    SensorName.HOSTNAME,
                    SensorKey.HOSTNAME,
                ),
            ],
        ),
        SensorCommand(
            "uname -m",
            sensors=[
                TextSensor(
                    SensorName.MACHINE_TYPE,
                    SensorKey.MACHINE_TYPE,
                ),
            ],
        ),
        SensorCommand(
            "uname -s",
            sensors=[
                TextSensor(
                    SensorName.OS_NAME,
                    SensorKey.OS_NAME,
                ),
            ],
        ),
        SensorCommand(
            "uname -r",
            sensors=[
                TextSensor(
                    SensorName.OS_VERSION,
                    SensorKey.OS_VERSION,
                ),
            ],
        ),
        SensorCommand(
            "dmidecode | grep -A4 '^System Information' | awk -F\": \" '{"
            + "if($0~/Product Name:/){a=$2} "
            + "if($0~/Version:/){b=$2} "
            + "if($0~/Manufacturer:/){c=$2} "
            + "if($0~/Serial Number:/){d=$2}} "
            + "END{print a; print b; print c; print d}'",
            sensors=[
                TextSensor(
                    SensorName.DEVICE_NAME,
                    SensorKey.DEVICE_NAME,
                ),
                TextSensor(
                    SensorName.DEVICE_MODEL,
                    SensorKey.DEVICE_MODEL,
                ),
                TextSensor(
                    SensorName.MANUFACTURER,
                    SensorKey.MANUFACTURER,
                ),
                TextSensor(
                    SensorName.SERIAL_NUMBER,
                    SensorKey.SERIAL_NUMBER,
                ),
            ],
        ),
        SensorCommand(
            'cat /proc/cpuinfo | awk -F": " \'{'
            + "if($0~/^model name/){a=$2} "
            + "if($0~/^processor/){b=$2} "
            + "if($0~/^Hardware/){c=$2} "
            + "if($0~/^Model/){d=$2}} "
            + "END{print a; print b+1; print c; print d}' | "
            + r'sed -e "s/[[:space:]]\+/ /g"',
            sensors=[
                TextSensor(
                    SensorName.CPU_NAME,
                    SensorKey.CPU_NAME,
                ),
                NumberSensor(
                    SensorName.CPU_CORES,
                    SensorKey.CPU_CORES,
                ),
                TextSensor(
                    SensorName.CPU_HARDWARE,
                    SensorKey.CPU_HARDWARE,
                ),
                TextSensor(
                    SensorName.CPU_MODEL,
                    SensorKey.CPU_MODEL,
                ),
            ],
        ),
        SensorCommand(
            "free -k | awk '/^Mem:/ {print $2}'",
            sensors=[
                NumberSensor(
                    SensorName.TOTAL_MEMORY,
                    SensorKey.TOTAL_MEMORY,
                    unit="KiB",
                )
            ],
        ),
        SensorCommand(
            "free -k | awk '/^Mem:/ {print $4}'",
            interval=30,
            sensors=[
                NumberSensor(
                    SensorName.FREE_MEMORY,
                    SensorKey.FREE_MEMORY,
                    unit="KiB",
                )
            ],
        ),
        SensorCommand(
            "df -k | "
            + r"awk '/^\/dev\// {"
            + "x=$4; "
            + '$1=$2=$3=$4=$5=""; '
            + 'sub(/^ +/, "", $0); '
            + 'print $0 "|" x}\'',
            interval=60,
            sensors=[
                NumberSensor(
                    SensorName.FREE_DISK_SPACE,
                    SensorKey.FREE_DISK_SPACE,
                    dynamic=True,
                    separator="|",
                    unit="KiB",
                )
            ],
        ),
        SensorCommand(
            "top -bn1 | awk 'NR<4 && tolower($0) ~ /cpu/ {print 100-$8}'",
            interval=30,
            sensors=[
                NumberSensor(
                    SensorName.CPU_LOAD,
                    SensorKey.CPU_LOAD,
                    unit="%",
                )
            ],
        ),
        SensorCommand(
            "for x in $(ls -d /sys/class/thermal/thermal_zone*); do "
            + "echo $(cat $x/type),$(($(cat $x/temp)/1000)); done",
            interval=60,
            sensors=[
                NumberSensor(
                    SensorName.TEMPERATURE,
                    SensorKey.TEMPERATURE,
                    dynamic=True,
                    separator=",",
                    unit="°C",
                )
            ],
        ),
        SensorCommand(
            "ps -e | awk 'END {print NR-1}'",
            interval=60,
            sensors=[
                NumberSensor(
                    SensorName.PROCESSES,
                    SensorKey.PROCESSES,
                )
            ],
        ),
    ],
)



from copy import deepcopy

from .command import ActionCommand, SensorCommand
from .sensor import Sensor


class Collection:
    action_commands: list[ActionCommand]
    sensor_commands: list[SensorCommand]

    def __init__(
        self,
        name: str,
        action_commands: list[ActionCommand] | None = None,
        sensor_commands: list[SensorCommand] | None = None,
    ) -> None:
        self.name = name
        self.set_action_commands(action_commands or [])
        self.set_sensor_commands(sensor_commands or [])

    @property
    def action_commands_by_key(self) -> dict[str, ActionCommand]:
        return {command.key: command for command in self.action_commands}

    @property
    def sensor_commands_by_sensor_key(self) -> dict[str, SensorCommand]:
        return {
            key: command
            for command in self.sensor_commands
            for key in command.sensors_by_key
        }

    @property
    def sensors_by_key(self) -> dict[str, Sensor]:
        return {
            key: sensor
            for command in self.sensor_commands
            for key, sensor in command.sensors_by_key.items()
        }

    def set_action_commands(self, action_commands: list[ActionCommand]) -> None:
        """Set the action commands."""
        self.action_commands = []
        for command in action_commands:
            self.add_action_command(command)

    def set_sensor_commands(self, sensor_commands: list[SensorCommand]) -> None:
        """Set the sensor commands."""
        self.sensor_commands = []
        for command in sensor_commands:
            self.add_sensor_command(command)

    def add_action_command(self, command: ActionCommand) -> None:
        """Add an action command.

        Remove existing action command with the same key.
        """
        if command.key in self.action_commands_by_key:
            self.remove_action_command(command.key)

        self.action_commands.append(deepcopy(command))

    def add_sensor_command(self, command: SensorCommand) -> None:
        """Add a sensor command.

        Remove existing sensors with the same keys.
        """
        for sensor in command.sensors:
            if sensor.key in self.sensors_by_key:
                self.remove_sensor(sensor.key)

        self.sensor_commands.append(deepcopy(command))

    def get_action_command(self, key: str) -> ActionCommand:
        """Get an action command."""
        return self.action_commands_by_key[key]

    def get_sensor_command(self, key: str) -> SensorCommand:
        """Get a sensor command."""
        return self.sensor_commands_by_sensor_key[key]

    def get_sensor(self, key: str) -> Sensor:
        """Get a sensor."""
        return self.sensors_by_key[key]

    def remove_action_command(self, key: str) -> None:
        """Remove an action command."""
        command = self.get_action_command(key)
        self.action_commands.remove(command)

    def remove_sensor(self, key: str) -> None:
        """Remove a sensor.

        Remove the sensor command if it doesnt have any other sensors.
        """
        command = self.get_sensor_command(key)
        command.remove_sensor(key)

        if not command.sensors_by_key:
            self.sensor_commands.remove(command)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import KW_ONLY, dataclass, field, replace
import re
from typing import TYPE_CHECKING, Any

from .event import Event
from .helpers import name_to_key

if TYPE_CHECKING:
    from . import Manager
    from .command import Command

TRUE_STRINGS = ["true", "enabled", "on", "active", "1"]
FALSE_STRINGS = ["false", "disabled", "off", "inactive", "0"]


@dataclass
class Sensor:
    name: str | None = None
    key: str | None = None
    _: KW_ONLY
    dynamic: bool = False
    separator: str | None = None
    unit: str | None = None
    renderer: Callable[[str], str] | None = None
    command_set: Command | None = None
    attributes: dict = field(default_factory=dict)

    def __post_init__(self):
        self.id = None
        self.key = self.key or name_to_key(self.name)
        self.value: Any | None = None
        self.last_known_value: Any | None = None
        self.child_sensors: list[Sensor] = []
        self.on_update = Event()
        self.on_child_add = Event()
        self.on_child_remove = Event()

    @property
    def controllable(self) -> bool:
        return self.command_set is not None

    @property
    def child_sensors_by_key(self) -> dict[str, Sensor]:
        return {child.key: child for child in self.child_sensors}

    def _get_control_command(self, _: Any) -> Command | None:
        return self.command_set

    def _add_child(self, child: Sensor) -> None:
        self.child_sensors.append(child)
        self.on_child_add.notify(self, child)

    def _remove_child(self, child: Sensor) -> None:
        self.child_sensors.remove(child)
        self.on_child_remove.notify(self, child)

    def _render(self, data: str) -> str:
        if self.renderer:
            data = self.renderer(data)

        return data.strip()

    def _convert(self, value_string: str) -> Any:
        return value_string

    def _validate(self, value: Any) -> None: ...

    def _update_value(self, manager: Manager, data: str | None) -> None:
        if data is None:
            self.value = None
            manager.logger.debug("%s: %s => %s", manager.name, self.key, self.value)
            return

        try:
            value_string = self._render(data)
            value = self._convert(value_string)
            self._validate(value)
        except Exception as exc:  # noqa: BLE001
            self.value = None
            manager.logger.debug(
                "%s: %s => %s (%s)", manager.name, self.key, self.value, exc
            )
            return

        self.value = self.last_known_value = value
        manager.logger.debug("%s: %s => %s", manager.name, self.key, self.value)

    def _update_child_sensors(self, manager: Manager, data: list[str] | None) -> None:
        if data is None:
            for child in self.child_sensors:
                child.update(manager, None)
            return

        dynamic_data_list = [
            DynamicData(self.name, self.key, *fields)
            for fields in (line.split(self.separator, 2) for line in data)
            if len(fields) >= 2
        ]

        dynamic_data_by_key = {
            dynamic_data.key: dynamic_data for dynamic_data in dynamic_data_list
        }

        for key, dynamic_data in dynamic_data_by_key.items():
            if key not in self.child_sensors_by_key:
                child = replace(
                    self,
                    name=dynamic_data.name,
                    key=dynamic_data.key,
                    dynamic=False,
                    separator=None,
                )
                child.id = dynamic_data.id
                self._add_child(child)

        for child in self.child_sensors:
            if child.key in dynamic_data_by_key:
                dynamic_data = dynamic_data_by_key[child.key]
                child.update(manager, dynamic_data.data)
            else:
                self._remove_child(child)

    def update(self, manager: Manager, data: Any) -> None:
        """Update and notify `on_update` subscribers."""
        if self.dynamic:
            self.value = self.last_known_value = None
            self._update_child_sensors(manager, data)
        else:
            self.child_sensors = []
            self._update_value(manager, data)

        self.on_update.notify(self)

    async def async_set(self, manager: Manager, value: Any) -> None:
        """Set a value.

        Raises:
            TypeError
            ValueError
            CommandError

        """
        self._validate(value)
        command = self._get_control_command(value)

        if command is None or value == self.value:
            return

        await manager.async_execute_command(
            command, variables={"id": self.id, "value": value}
        )


@dataclass
class TextSensor(Sensor):
    _: KW_ONLY
    minimum: int | None = None
    maximum: int | None = None
    pattern: str | None = None
    options: list[Any] | None = None

    def _convert(self, value_string: str) -> str:
        if value_string == "":
            raise ValueError("String is empty")

        return value_string

    def _validate(self, value: Any) -> None:
        if not isinstance(value, str):
            raise TypeError(f"{value} is {type(value)} and not {str}")

        if self.minimum and len(value) < self.minimum:
            raise ValueError(f"{value} is shorter then {self.minimum}")

        if self.maximum and len(value) > self.maximum:
            raise ValueError(f"{value} is longer then {self.maximum}")

        if self.pattern and not re.fullmatch(self.pattern, value):
            raise ValueError(f"{value} doesn't match {self.pattern}")

        if self.options and value not in self.options:
            raise ValueError(f"{value} is not in {self.options}")


@dataclass
class NumberSensor(Sensor):
    _: KW_ONLY
    float: bool = False
    minimum: int | float | None = None
    maximum: int | float | None = None

    def _convert(self, value_string: str) -> int | float:
        if self.float:
            return float(value_string)

        return int(float(value_string))

    def _validate(self, value: Any) -> None:
        if self.float and not isinstance(value, float):
            raise TypeError(f"{value} is {type(value)} and not {float}")

        if not self.float and not isinstance(value, int):
            raise TypeError(f"{value} is {type(value)} and not {int}")

        if self.minimum and value < self.minimum:
            raise ValueError(f"{value} is smaller then {self.minimum}")

        if self.maximum and value > self.maximum:
            raise ValueError(f"{value} is bigger then {self.maximum}")


@dataclass
class BinarySensor(Sensor):
    _: KW_ONLY
    command_on: Command | None = None
    command_off: Command | None = None
    payload_on: str | None = None
    payload_off: str | None = None

    @property
    def controllable(self) -> bool:
        if self.command_on and self.command_off:
            return True

        return super().controllable

    def _get_control_command(self, value: Any) -> Command | None:
        if self.command_on and value is True:
            return self.command_on

        if self.command_off and value is False:
            return self.command_off

        return self.command_set

    def _convert(self, value_string: str) -> bool:
        if self.payload_on:
            if value_string == self.payload_on:
                return True
            if not self.payload_off:
                return False

        if self.payload_off:
            if value_string == self.payload_off:
                return False
            if not self.payload_on:
                return True

        if value_string.lower() in TRUE_STRINGS:
            return True

        if value_string.lower() in FALSE_STRINGS:
            return False

        raise ValueError(f"Can't generate bool from {value_string}")

    def _validate(self, value: Any) -> None:
        if not isinstance(value, bool):
            raise TypeError(f"{value} is {type(value)} and not {bool}")


class DynamicData:
    def __init__(
        self,
        parent_name: str | None,
        parent_key: str,
        id_field: str,
        data_field: str,
        name_field: str | None = None,
    ) -> None:
        self.id = id_field.strip()
        name = name_field.strip() if name_field else self.id
        self.key = f"{parent_key}_{name_to_key(name)}"
        self.name = f"{parent_name} {name}" if parent_name else name
        self.data = data_field



from __future__ import annotations

from collections.abc import Callable
from dataclasses import KW_ONLY, dataclass, field
import re as _re
from string import Template
from time import time
from typing import TYPE_CHECKING

from .errors import CommandError
from .helpers import name_to_key
from .sensor import Sensor

if TYPE_CHECKING:
    from . import CommandOutput, Manager

SENSOR_DELIMITER = "&"
VARIABLE_DELIMITER = "@"
PLACEHOLDER_KEY = "_"


# Not needed anymore after python 3.11
class Template(Template):
    def get_identifiers(self):
        ids = []
        for mo in self.pattern.finditer(self.template):
            named = mo.group("named") or mo.group("braced")
            if named is not None and named not in ids:
                ids.append(named)
            elif (
                named is None
                and mo.group("invalid") is None
                and mo.group("escaped") is None
            ):
                raise ValueError("Unrecognized named group in pattern", self.pattern)
        return ids


class SensorTemplate(Template):
    delimiter = SENSOR_DELIMITER
    pattern = rf"""
    {_re.escape(delimiter)}{{(?:
      (?P<braced>{Template.idpattern})|
      (?P<invalid>)|
      (?P<named>)|
      (?P<escaped>)
    )}}
    """


class VariableTemplate(Template):
    delimiter = VARIABLE_DELIMITER
    pattern = rf"""
    {_re.escape(delimiter)}{{(?:
      (?P<braced>{Template.idpattern})|
      (?P<invalid>)|
      (?P<named>)|
      (?P<escaped>)
    )}}
    """


@dataclass
class Command:
    string: str
    _: KW_ONLY
    timeout: float | None = None
    renderer: Callable[[str], str] | None = None

    @property
    def required_variables(self) -> set[str]:
        return set(VariableTemplate(self.string).get_identifiers())

    @property
    def required_sensors(self) -> set[str]:
        return set(SensorTemplate(self.string).get_identifiers())

    def _render(self, string: str) -> str:
        if self.renderer:
            string = self.renderer(string)
        return string

    async def async_generate_string(
        self,
        manager: Manager,
        variables: dict | None = None,
    ) -> str:
        """Generate the string.

        Raises:
            CommandError

        """
        variables = variables or {}
        sensor_values_by_key = {}
        string = self.string

        try:
            string = VariableTemplate(string).substitute(variables)
        except Exception as exc:  # noqa: BLE001
            raise CommandError(f"Failed to substitute variable ({exc})") from exc

        try:
            sensors = await manager.async_poll_sensors(self.required_sensors)
        except Exception as exc:  # noqa: BLE001
            raise CommandError(f"Failed to poll sensors ({exc})") from exc

        for sensor in sensors:
            if sensor.value is not None:
                sensor_values_by_key[sensor.key] = sensor.value
            else:
                raise CommandError(f"Value of sensor {sensor.key} is None")

        try:
            string = SensorTemplate(string).substitute(sensor_values_by_key)
        except Exception as exc:
            raise CommandError(f"Failed to substitute sensor ({exc})") from exc

        try:
            return self._render(string)
        except Exception as exc:  # noqa: BLE001
            raise CommandError(f"Failed to render string ({exc})") from exc

    async def async_execute(
        self, manager: Manager, variables: dict | None = None
    ) -> CommandOutput:
        """Execute.

        Raises:
            CommandError

        """
        try:
            string = await self.async_generate_string(manager, variables)
        except CommandError as exc:
            manager.logger.debug("%s: %s => %s", manager.name, self.string, exc)
            raise

        try:
            output = await manager.async_execute_command_string(string, self.timeout)
        except CommandError as exc:
            manager.logger.debug("%s: %s => %s", manager.name, string, exc)
            raise

        manager.logger.debug(
            "%s: %s => %s, %s, %s",
            manager.name,
            string,
            output.stdout,
            output.stderr,
            output.code,
        )

        return output


@dataclass
class ActionCommand(Command):
    name: str | None = None
    key: str | None = None
    _: KW_ONLY
    attributes: dict = field(default_factory=dict)

    def __post_init__(self):
        self.key = self.key or name_to_key(self.name)


@dataclass
class SensorCommand(Command):
    _: KW_ONLY
    interval: int | None = None
    sensors: list[Sensor] = field(default_factory=list)

    def __post_init__(self):
        self.last_update: float | None = None

    @property
    def should_update(self) -> bool:
        if not self.interval:
            return False
        if not self.last_update:
            return True
        if time() - self.last_update < self.interval:
            return False
        return True

    @property
    def sensors_by_key(self) -> dict[str, Sensor]:
        return {
            sensor.key: sensor
            for command_sensor in self.sensors
            for sensor in (command_sensor, *command_sensor.child_sensors)
            if command_sensor.key != PLACEHOLDER_KEY
        }

    def remove_sensor(self, key: str) -> None:
        """Remove a sensor."""
        self.sensors = [
            Sensor(key=PLACEHOLDER_KEY) if sensor.key == key else sensor
            for sensor in self.sensors
        ]

    def update_sensors(self, manager: Manager, output: CommandOutput | None) -> None:
        """Update the sensors."""
        if output and output.code == 0:
            self.last_update = output.timestamp
            data = output.stdout
        else:
            data = []

        for i, sensor in enumerate(self.sensors):
            if sensor.dynamic:
                sensor.update(manager, data[i:] or None)
                return
            sensor.update(manager, data[i] if len(data) > i else None)

    async def async_execute(
        self, manager: Manager, variables: dict | None = None
    ) -> CommandOutput:
        try:
            output = await super().async_execute(manager, variables)
        except CommandError:
            self.update_sensors(manager, None)
            raise

        self.update_sensors(manager, output)

        return output



from __future__ import annotations

import asyncio
from collections.abc import Coroutine
from functools import wraps
import inspect


def locked(coro: Coroutine):
    @wraps(coro)
    async def wrapper(obj: Synchronizer, *args, **kwargs):
        async with obj.lock:
            return await coro(obj, *args, **kwargs)

    return wrapper


class AsyncRLock(asyncio.Lock):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._task = None
        self._depth = 0

    async def acquire(self) -> None:
        if self._task is None or self._task != asyncio.current_task():
            await super().acquire()
            self._task = asyncio.current_task()
            assert self._depth == 0
        self._depth += 1

    def release(self) -> None:
        if self._depth > 0:
            self._depth -= 1
        if self._depth == 0:
            super().release()
            self._task = None


class Synchronizer:
    def __init__(self) -> None:
        self.lock = AsyncRLock()

    def __init_subclass__(cls, **kwargs):
        for name in cls.__dict__:
            attr = getattr(cls, name)
            if inspect.iscoroutinefunction(attr):
                setattr(cls, name, locked(attr))



from collections.abc import Callable


class Event:
    def __init__(self) -> None:
        self._subscribers: list[Callable] = []

    def subscribe(self, subscriber) -> Callable:
        """Subscribe."""
        self._subscribers.append(subscriber)

        def callback():
            self.unsubscribe(subscriber)

        return callback

    def unsubscribe(self, subscriber) -> None:
        """Unsubscribe."""
        self._subscribers.remove(subscriber)

    def notify(self, *args, **kwargs) -> None:
        """Notify."""
        for subscriber in self._subscribers:
            subscriber(*args, **kwargs)
